/*
 * Copyright (C) 2022 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.operator.keycloak.user;

import static org.assertj.core.api.Assertions.*;
import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.Collections;
import java.util.List;
import java.util.Optional;

import jakarta.ws.rs.core.Response;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.keycloak.admin.client.Keycloak;
import org.keycloak.admin.client.resource.ClientsResource;
import org.keycloak.admin.client.resource.RealmResource;
import org.keycloak.admin.client.resource.RoleMappingResource;
import org.keycloak.admin.client.resource.RoleScopeResource;
import org.keycloak.admin.client.resource.UserResource;
import org.keycloak.admin.client.resource.UsersResource;
import org.keycloak.representations.idm.ClientRepresentation;
import org.keycloak.representations.idm.RoleRepresentation;
import org.keycloak.representations.idm.UserRepresentation;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;

import de.ozgcloud.operator.keycloak.KeycloakException;
import de.ozgcloud.operator.keycloak.KeycloakGenericRemoteService;

class KeycloakUserRemoteServiceTest {

	private final static UserRepresentation USER_REPRESENTATION = UserRepresentationTestFactory.create();
	private final static String REALM = "TestRealm";
	private final static String USERID = "UserId";
	private final static String CLIENT_ID = "ClientId";
	private final static String CLIENT_ROLE_NAME = "ClientRoleName";

	@Spy
	@InjectMocks
	private KeycloakUserRemoteService userRemoteService;

	@Mock
	private KeycloakGenericRemoteService genericRemoteService;

	@Mock
	private Keycloak keycloak;
	@Mock
	private RealmResource realmResource;
	@Mock
	private UsersResource usersResource;
	@Mock
	private UserResource userResource;
	@Mock
	private ClientsResource clientsResource;
	@Mock
	private ClientRepresentation clientRepresentation;
	@Mock
	private RoleRepresentation roleRepresentation;
	@Mock
	private RoleScopeResource roleScopeResource;
	@Mock
	private RoleMappingResource roleMappingResource;
	@Mock
	private UserRepresentation userRepresentation;
	@Mock
	private KubernetesRemoteService kubernetesRemoteService;

	@Nested
	class TestCreateUser {

		@Mock
		private Response response;

		@BeforeEach
		void init() {
			when(keycloak.realm(REALM)).thenReturn(realmResource);
			when(realmResource.users()).thenReturn(usersResource);
			when(usersResource.create(USER_REPRESENTATION)).thenReturn(response);
			when(response.getStatusInfo()).thenReturn(Response.Status.CREATED);
		}

		@Test
		void shouldThrowOnResponseError() {
			when(response.getStatusInfo()).thenReturn(Response.Status.INTERNAL_SERVER_ERROR);

			assertThrows(KeycloakException.class, () -> userRemoteService.createUser(USER_REPRESENTATION, REALM));
		}

		@Test
		void shouldCallAddClientRoles() {
			when(response.getStatusInfo()).thenReturn(Response.Status.CREATED);
			doNothing().when(userRemoteService).addClientRoles(any(), any(), eq(REALM), any());

			userRemoteService.createUser(USER_REPRESENTATION, REALM);

			verify(userRemoteService).addClientRoles(isNull(), eq(realmResource), eq(REALM), eq(USER_REPRESENTATION));
		}
	}

	@Nested
	class TestGetRealmClient {

		@BeforeEach
		void init() {
			when(realmResource.clients()).thenReturn(clientsResource);
		}

		@Test
		void shouldReturnClientRespresentation() {
			when(clientsResource.findByClientId(UserRepresentationTestFactory.CLIENT_NAME)).thenReturn(List.of(clientRepresentation));

			ClientRepresentation newClientRepresentation = userRemoteService.getRealmClient(realmResource, UserRepresentationTestFactory.CLIENT_NAME);

			assertThat(newClientRepresentation).isSameAs(clientRepresentation);
		}

		@Test
		void shouldThrowOnMissingClient() {
			when(clientsResource.findByClientId(UserRepresentationTestFactory.CLIENT_NAME)).thenReturn(Collections.emptyList());

			assertThrows(KeycloakException.class, () -> userRemoteService.getRealmClient(realmResource, UserRepresentationTestFactory.CLIENT_NAME));
		}
	}

	@Nested
	class TestAddClientRoles {

		@BeforeEach
		void init() {
			when(clientRepresentation.getId()).thenReturn(CLIENT_ID);
			doReturn(clientRepresentation).when(userRemoteService).getRealmClient(any(), any());
		}

		@Test
		void shouldAddClientRole() {
			var user = UserRepresentationTestFactory.create();
			when(genericRemoteService.getClientRole(UserRepresentationTestFactory.ROLE1, CLIENT_ID, REALM))
					.thenReturn(Optional.of(roleRepresentation));
			doNothing().when(userRemoteService).addClientRoleToUser(any(), any(), any(), any());

			userRemoteService.addClientRoles(USERID, realmResource, REALM, user);

			verify(userRemoteService).addClientRoleToUser(roleRepresentation, realmResource, USERID, clientRepresentation);
		}

		@Test
		void shouldThrowOnMissingClientRole() {
			var user = UserRepresentationTestFactory.create();
			when(genericRemoteService.getClientRole(any(), any(), any())).thenReturn(Optional.empty());

			assertThrows(KeycloakException.class,
					() -> userRemoteService.addClientRoles(USERID, realmResource, REALM, user));
		}
	}

	@Nested
	class TestAddClientRolesUserWithoutRoles {

		@Test
		void shouldNotThrowOnUserWithoutRoles() {
			UserRepresentation user = createUserWithNullRoles();

			assertDoesNotThrow(() -> userRemoteService.addClientRoles(USERID, realmResource, REALM, user));
		}

		private UserRepresentation createUserWithNullRoles() {
			UserRepresentation user = UserRepresentationTestFactory.create();
			user.setClientRoles(null);
			return user;
		}
	}

	@Nested
	class TestGetRealmClientNamesOfUser {

		@Test
		void shouldReturnClientNames() {
			UserRepresentation user = UserRepresentationTestFactory.create();

			List<String> clientNames = userRemoteService.getClientNamesOfUser(user).toList();

			assertThat(clientNames).contains(UserRepresentationTestFactory.CLIENT_NAME);
		}
	}

	@Nested
	class TestGetClientRole {

		@Spy
		@InjectMocks
		private KeycloakUserRemoteService userRemoteService;

		@Mock
		private KeycloakGenericRemoteService keycloakGenericRemoteService;

		@BeforeEach
		void init() {
			when(clientRepresentation.getId()).thenReturn(CLIENT_ID);
		}

		@Test
		void shouldReturnClientRole() {
			when(keycloakGenericRemoteService.getClientRole(CLIENT_ROLE_NAME, CLIENT_ID, REALM)).thenReturn(Optional.of(roleRepresentation));

			RoleRepresentation role = userRemoteService.getClientRole(REALM, CLIENT_ID, clientRepresentation, CLIENT_ROLE_NAME);

			assertThat(role).isEqualTo(roleRepresentation);
		}

		@Test
		void shouldThrowOnMissingRole() {
			when(keycloakGenericRemoteService.getClientRole(CLIENT_ROLE_NAME, CLIENT_ID, REALM)).thenReturn(Optional.empty());

			assertThrows(KeycloakException.class,
					() -> userRemoteService.getClientRole(REALM, CLIENT_ID, clientRepresentation, CLIENT_ROLE_NAME));
		}
	}

	@Nested
	class TestAddClientRoleToUser {

		@BeforeEach
		void init() {
			when(realmResource.users()).thenReturn(usersResource);
			when(usersResource.get(USERID)).thenReturn(userResource);
			when(userResource.roles()).thenReturn(roleMappingResource);
			when(roleMappingResource.clientLevel(CLIENT_ID)).thenReturn(roleScopeResource);

			when(clientRepresentation.getId()).thenReturn(CLIENT_ID);
		}

		@Test
		void shouldAddClientRoleToUser() {
			userRemoteService.addClientRoleToUser(roleRepresentation, realmResource, USERID, clientRepresentation);

			verify(roleScopeResource).add(List.of(roleRepresentation));
		}
	}

	@DisplayName("Update user")
	@Nested
	class TestUpdateUser {

		private final static String USERID = "UserId";

		@BeforeEach
		void init() {
			when(keycloak.realm(REALM)).thenReturn(realmResource);
			when(realmResource.users()).thenReturn(usersResource);
			when(usersResource.get(any())).thenReturn(userResource);

			doNothing().when(userRemoteService).addClientRoles(any(), any(), eq(REALM), any());
		}

		@Test
		void shouldCallKeycloakApiUpdateUser() {
			var user = UserRepresentationTestFactory.create();
			user.setId(USERID);

			userRemoteService.updateUser(user, REALM);

			verify(userResource).update(user);
		}

		@Test
		void shouldCallAddClientRoles() {
			userRemoteService.updateUser(USER_REPRESENTATION, REALM);

			verify(userRemoteService).addClientRoles(isNull(), eq(realmResource), eq(REALM), eq(USER_REPRESENTATION));
		}
	}

	@Nested
	class TestGetUserByName {

		@BeforeEach
		void init() {
			var userList = List.of(UserRepresentationTestFactory.create());

			when(keycloak.realm(REALM)).thenReturn(realmResource);
			when(realmResource.users()).thenReturn(usersResource);
			when(usersResource.search(USER_REPRESENTATION.getUsername(), true)).thenReturn(userList);
		}

		@Test
		void shouldGetUserFromKeycloak() {

			Optional<UserRepresentation> user = userRemoteService.getUserByName(UserRepresentationTestFactory.USERNAME, REALM);

			assertThat(user).isPresent();
		}
	}

	@Nested
	class TestDeleteUser {

		@BeforeEach
		void init() {
			when(keycloak.realm(REALM)).thenReturn(realmResource);
			when(realmResource.users()).thenReturn(usersResource);
		}

		@Test
		void shouldDeleteUser() {
			userRemoteService.deleteUser(USERID, REALM);

			verify(usersResource).delete(USERID);
		}
	}
}